use std::{
    path::PathBuf,
    sync::{Arc, Mutex},
};

use clap::Command;

use crate::{context::CuddleContext, model::CuddleScript};

#[derive(Debug, Clone)]
#[allow(dead_code)]
struct CuddleAction {
    script: CuddleScript,
    path: PathBuf,
    name: String,
}
#[allow(dead_code)]
impl CuddleAction {
    pub fn new(script: CuddleScript, path: PathBuf, name: String) -> Self {
        Self { script, path, name }
    }

    pub fn execute(self) {
        match self.script {
            CuddleScript::Shell(s) => {}
            CuddleScript::Dagger(d) => {}
        }
    }
}

#[derive(Debug, Clone)]
pub struct CuddleCli<'a> {
    scripts: Vec<CuddleAction>,
    context: Arc<Mutex<Vec<CuddleContext>>>,
    command: Option<Command<'a>>,
}

impl<'a> CuddleCli<'a> {
    pub fn new(context: Arc<Mutex<Vec<CuddleContext>>>) -> anyhow::Result<CuddleCli<'a>> {
        let mut cli = CuddleCli {
            scripts: vec![],
            context: context.clone(),
            command: None,
        };

        cli = cli.process_scripts().build_cli();

        Ok(cli)
    }

    fn process_scripts(mut self) -> Self {
        if let Ok(context_iter) = self.context.clone().lock() {
            for ctx in context_iter.iter() {
                if let Some(scripts) = ctx.plan.scripts.clone() {
                    for (name, script) in scripts {
                        self.scripts
                            .push(CuddleAction::new(script.clone(), ctx.path.clone(), name))
                    }
                }
            }
        }

        self
    }

    fn build_cli(mut self) -> Self {
        let mut root_cmd = Command::new("cuddle")
            .version("1.0")
            .author("kjuulh <contact@kasperhermansen.com>")
            .about("cuddle is your domain specific organization tool. It enabled widespread sharing through repositories, as well as collaborating while maintaining speed and integrity")
            .propagate_version(true)
            .arg_required_else_help(true);

        let mut execute_cmd = Command::new("x").about("x is your entry into your domains scripts, scripts inherited from parents will also be present here");

        for script in self.scripts.iter() {
            let action_cmd = Command::new(script.name.clone());

            // TODO: Some way to add an about for clap, requires conversion from String -> &str
            execute_cmd = execute_cmd.subcommand(action_cmd);
        }

        root_cmd = root_cmd.subcommand(execute_cmd);

        self.command = Some(root_cmd);

        self
    }

    pub fn execute(self) -> Self {
        if let Some(cli) = self.command.clone() {
            let _ = cli.get_matches();
        }

        self
    }
}
